// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef FILEBUF__HPP
#define FILEBUF__HPP

#include <ds_basicstreambuf.hpp>
#include <stdio.h>

namespace DS_NAMESPACE
{

  template <TYPENAME charT>
  class basic_filebuf : public basic_streambuf<charT>
  {
  public:
    typedef typename basic_streambuf<charT>::char_type char_type;
    typedef typename basic_streambuf<charT>::int_type int_type;
    typedef typename basic_streambuf<charT>::pos_type pos_type;
    typedef typename basic_streambuf<charT>::off_type off_type;

    basic_filebuf() : fp_(0) { }
    ~basic_filebuf();

    bool is_open() const { return fp_; }

    basic_filebuf *open(const char *filename, ios_base::openmode mode)
    {
      if(!filename)
	return 0;

      char buf[4]; buf[0]=buf[1]=buf[2]=buf[3]=0;
      char *bp=buf;
      switch(mode & ~(ios::ate|ios::binary))
	{
	case ios::in : *bp++='r'; break;
	case ios::out :
	case ios::out|ios::trunc : *bp++='w'; break;
	case ios::out|ios::app   : *bp++='a'; break;
	case ios::out|ios::in    : *bp++='r'; *bp++='+'; break;
	case ios::out|ios::in|ios::trunc : *bp++='w'; *bp++='+'; break;
	case ios::out|ios::in|ios::app   : *bp++='a'; *bp++='+'; break;
	default: return 0;
	}
      if(mode & ios::binary)
	*bp++='b';

      close();
      fp_=fopen(filename, buf);
      if(!fp_)
	return 0;

      if(mode & ios::ate)
	if(fseek(fp_, 0, SEEK_END) < 0)
	  close();

      return fp_?this:0;
    }

    basic_filebuf *close()
    {
      if(!fp_)
	return 0;
      const int ret=fclose(fp_);
      fp_=0;
      return ret?0:this;
    }

    basic_filebuf *open(FILE *f)
    {
      close();
      fp_=f;
      return fp_?this:0;
    }

  protected:
    virtual pos_type seekoff(off_type off, ios_base::seekdir way, ios_base::openmode which = ios_base::in | ios_base::out)
    {
      if(!fp_)
	return EOF;
      int s=0;
      switch(way)
	{
	case ios::beg: s=SEEK_SET; break;
	case ios::cur: s=SEEK_CUR; break;
	case ios::end: s=SEEK_END; break;
	default: return EOF;
	}
      if(fseek(fp_, off, s) < 0)
	return EOF;
      fpos_t p;
      if(fgetpos(fp_, &p) < 0)
	return EOF;
#if defined(_WIN32_WCE)
      return static_cast<pos_type>(p);
#else
      return static_cast<pos_type>(p.__pos);
#endif
    }

    virtual pos_type seekpos(pos_type pos, ios_base::openmode which = ios_base::in | ios_base::out)
    {
      if(!fp_)
	return EOF;
      fpos_t p;
#if defined(_WIN32_WCE)
      p=pos;
#else
      p.__pos=pos;
#endif
      if(fsetpos(fp_, &p) < 0)
	return EOF;
#if defined(_WIN32_WCE)
      return static_cast<pos_type>(p);
#else
      return p.__pos;
#endif
    }

    virtual int_type underflow()
    {
      if(!fp_)
	return EOF;
      if(feof(fp_) || ferror(fp_))
	return EOF;
      char_type c;
      if(fread(&c, sizeof(char_type), 1, fp_) != 1)
	return EOF;
      return c;
    }

    virtual int_type pbackfail(int_type meta = EOF)
    {
      if(fseek(fp_, -static_cast<int>(sizeof(char_type)), SEEK_CUR) < 0)
	return EOF;
      return !EOF;
    }

    virtual int_type overflow(int_type meta = EOF)
    {
      if(meta!=EOF)
	{
	  char_type c=meta;
	  if(fwrite(&c, sizeof(char_type), 1, fp_) != 1)
	    return EOF;
	  return meta;
	}
      return EOF;
    }

    virtual basic_streambuf<charT> *setbuf(charT *buffer, streamsize count)
    {
      if(!fp_)
	return 0;
#if !defined(_WIN32_WCE)
      if(setvbuf(fp_, (char*)buffer, _IOFBF, count*sizeof(charT)))
	return 0;
#endif
      return this;
    }

    virtual int sync()
    {
      if(!fp_)
	return 0;
      if(this->overflow()!=EOF && fflush(fp_)==0)
	return 0;
      return EOF;
    }

    virtual streamsize xsgetn(charT *ptr, streamsize count);
    virtual streamsize xsputn(const charT *ptr, streamsize count);


  private:
    FILE *fp_;
  };

  template <TYPENAME charT>
  basic_filebuf<charT>::~basic_filebuf()
  {
    close();
  }

  template <TYPENAME charT>
  streamsize basic_filebuf<charT>::xsgetn(charT *ptr, streamsize count)
  {
    if(feof(fp_) || ferror(fp_))
      return EOF;
    return fread(ptr, sizeof(charT), static_cast<size_t>(count), fp_);
  }

  template <TYPENAME charT>
  streamsize basic_filebuf<charT>::xsputn(const charT *ptr, streamsize count)
  {
    return fwrite(ptr, sizeof(charT), static_cast<size_t>(count), fp_);
  }

  typedef basic_filebuf<char> filebuf;
  typedef basic_filebuf<wchar_t> wfilebuf;
}

#endif
