// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef DS_BASICIOS__HPP
#define DS_BASICIOS__HPP

#include <ds_basicstreambuf.hpp>
#include <string>
#include <cstdio>

namespace DS_NAMESPACE
{
  extern char *truestr;
  extern char *falsestr;
  extern char *zerostr;
  extern wchar_t *truewstr;
  extern wchar_t *falsewstr;
  extern wchar_t *zerowstr;

  extern char *uppernumbers;
  extern char *lowernumbers;
  extern wchar_t *upperwnumbers;
  extern wchar_t *lowerwnumbers;

  template <TYPENAME charT>
  class basic_ios_traits
  {
  public:
    charT *truestring, *falsestring, *zerostring;
    charT *uppernumber, *lowernumber;
    basic_ios_traits() :
      truestring(0), falsestring(0), zerostring(0),
      uppernumber(0), lowernumber(0)
    { }
  };

  template <>
  class basic_ios_traits<char>
  {
  public:
    char *truestring, *falsestring, *zerostring;
    char *uppernumber, *lowernumber;
    basic_ios_traits() :
      truestring(truestr), falsestring(falsestr), zerostring(zerostr),
      uppernumber(uppernumbers), lowernumber(lowernumbers)
    { }
  };

  template <>
  class basic_ios_traits<wchar_t>
  {
  public:
    wchar_t *truestring, *falsestring, *zerostring;
    wchar_t *uppernumber, *lowernumber;
    basic_ios_traits() :
      truestring(truewstr), falsestring(falsewstr), zerostring(zerowstr),
      uppernumber(upperwnumbers), lowernumber(lowerwnumbers)
    { }
  };

  template <TYPENAME charT>
  class basic_ios : public ios_base
  {
  public:
    typedef basic_ios<charT>* tie_t;

  private:
    tie_t tie_;
    void flushtie();
    void sync();

  protected:
    charT fill_;

    typedef basic_streambuf<charT> sb_t;
    sb_t *sb_;
    int sbput(const int i);
    int sbget();

    void widen(charT *dst, char *src);
    void narrow_(char *dst, charT *src);

  public:
    basic_ios<charT>(basic_streambuf<charT> *s=0, tie_t t=0);
    ~basic_ios<charT>();

    basic_streambuf<charT>* rdbuf() const { return sb_; }
    basic_streambuf<charT>* rdbuf(basic_streambuf<charT> *newsb);

    tie_t tie() const { return tie_; }
    tie_t tie(tie_t n)
    {
      tie_t o=tie_;
      tie_=n;
      return o;
    }

    charT fill() const { return fill_; }
    charT fill(int f)
    {
      const charT oldfill_=fill_;
      fill_=f;
      return oldfill_;
    }

    iostate exceptions() const { return 0; }
    iostate exceptions(iostate newexcept) { return 0; }

    basic_ios<charT>& copyfmt(const basic_ios<charT>& i);

    void init(basic_streambuf<charT> *newsb);

    charT widen(const char c) { return static_cast<charT>(c); }
    char narrow(const int c, const char def) { return static_cast<char>(c); }

  private:
    basic_ios<charT> (const basic_ios<charT>&); // we don't want copy construction
    basic_ios<charT>& operator= (const basic_ios<charT>&); // we don't want assign operations
  };

  typedef basic_ios<char> ios;
  typedef basic_ios<wchar_t> wios;



  template <TYPENAME charT>
  void basic_ios<charT>::flushtie()
  {
    if(tie_ && sb_)
      tie_->sync();
  }

  template <TYPENAME charT>
  void basic_ios<charT>::sync()
  {
    if(sb_)
      sb_->pubsync();
  }

  template <TYPENAME charT>
  int basic_ios<charT>::sbput(const int i)
  {
    flushtie();
    if(!sb_)
      return EOF;
    return sb_->sputc(i);
  }

  template <TYPENAME charT>
  int basic_ios<charT>::sbget()
  {
    flushtie();
    if(!sb_)
      return EOF;
    return sb_->sbumpc();
  }

  template <TYPENAME charT>
  basic_ios<charT>::basic_ios(basic_streambuf<charT> *s, tie_t t) :
    tie_(t),
    sb_(0)
  {
    init(s);
  }

  template <TYPENAME charT>
  basic_ios<charT>::~basic_ios()
  {
  }

  template <TYPENAME charT>
  basic_streambuf<charT>* basic_ios<charT>::rdbuf(basic_streambuf<charT> *newsb)
  {
    sb_t *oldsb=sb_;
    sb_=newsb;
    if(sb_)
      setstate(goodbit);
    else
      setstate(badbit);
    return oldsb;
  }

  template <TYPENAME charT>
  basic_ios<charT>& basic_ios<charT>::copyfmt(const basic_ios<charT>& i)
  {
    // todo: this seems buggy
    fill_=i.fill_;
    fmtflags_=i.fmtflags_;
    rdbuf(i.rdbuf());
    return *this;
  }

  template <TYPENAME charT>
  void basic_ios<charT>::init(basic_streambuf<charT> *newsb)
  {
    rdbuf(newsb);
    tie(0);
    //exceptions(); has to return goodbit
    fmtflags_=static_cast<fmtflags>(skipws|dec);
    width_=0;
    precision_=6;
    fill_=' ';
  }

  template <TYPENAME charT>
  void basic_ios<charT>::widen(charT *dst, char *src)
  {
    while(*src)
      {
	*dst=static_cast<charT>(*src);
	++dst;
	++src;
      }
    *dst=static_cast<charT>(0);
  }

  template <TYPENAME charT>
  void basic_ios<charT>::narrow_(char *dst, charT *src)
  {
    while(*src)
      {
	*dst=static_cast<char>(*src);
	++dst;
	++src;
      }
    *dst=static_cast<char>(0);
  }

}

#endif
