// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef DS_BASICISTREAM__HPP
#define DS_BASICISTREAM__HPP

#include <ctype.h>
#include <iosfwd>
#include <string>
#include <ds_basicios.hpp>

namespace DS_NAMESPACE
{
  template <TYPENAME charT,
	    TYPENAME _traits
#if defined(__GNUC__)
	    = int
#endif
	    >
  class basic_istream : public virtual basic_ios<charT>
  {
    typedef basic_ios<charT> bios;
    using bios::sb_;
  protected:
    size_t count;

    int getImp();
    void skipwhitespace();

    bool isws(const int c);

  public:
    basic_istream<charT, _traits>(basic_streambuf<charT> *s=0) :
      basic_ios<charT>(s)
    {
      //ds_debughook("basic_istream()");
    }

    ~basic_istream<charT, _traits>()
    {
      //ds_debughook("~basic_istream()");
    }

    streampos tellg() const;
    basic_istream<charT, _traits>& seekg(const streamoff offs);
    basic_istream<charT, _traits>& seekg(const streamoff offs, const ios::seekdir flag);

    int sync() const;

    int get();
    basic_istream<charT, _traits>& get(charT& c);
    basic_istream<charT, _traits>& get(charT* zf, size_t anz, charT ende='\n');
    basic_istream<charT, _traits>& getline(charT *zf, size_t anz, charT ende='\n');
    basic_istream<charT, _traits>& read(charT *zf, size_t anz);
    basic_istream<charT, _traits>& readsome(charT *zf, size_t anz);

    size_t gcount() const { return count; }
    basic_istream<charT, _traits>& ignore(int anz=1, int ende=EOF);
    int peek();
    basic_istream<charT, _traits>& unget();
    basic_istream<charT, _traits>& putback(const charT c);
    basic_istream<charT, _traits>& get(basic_streambuf<charT>& s, const charT delim='\n');

    basic_istream<charT, _traits>& operator>> (bool& b);

    basic_istream<charT, _traits>& operator>> (unsigned int& num);
    basic_istream<charT, _traits>& operator>> (int& num);
    basic_istream<charT, _traits>& operator>> (charT *str);
    basic_istream<charT, _traits>& operator>> (DS_STRING_NAMESPACE::basic_string<charT>& str);
    basic_istream<charT, _traits>& operator>> (void* &v);
    basic_istream<charT, _traits>& operator>> (double&);

    basic_istream<charT, _traits>& operator>> (float& f)
    {
      double d; *this >> d; f=static_cast<float>(d); return *this;
    }

    // iomanip overloads

    basic_istream<charT, _traits>& operator>> (resetiosflags_t mask)
    {
      this->unsetf(mask.flags);
      return *this;
    }

    basic_istream<charT, _traits>& operator>> (setiosflags_t mask)
    {
      this->setf(mask.flags);
      return *this;
    }

    basic_istream<charT, _traits>& operator>> (setbase_t base)
    {
      switch(base.base)
	{
	case 8:  this->setf(oct, bios::basefield); break;
	case 10: this->setf(dec, bios::basefield); break;
	case 16: this->setf(hex, bios::basefield); break;
	default: this->setf(0,   bios::basefield); break;
	}
      return *this;
    }

    basic_istream<charT, _traits>& operator>> (setfill_t f)
    {
      this->fill(f.fill);
      return *this;
    }

    basic_istream<charT, _traits>& operator>> (setprecision_t p)
    {
      this->precision(p.prec);
      return *this;
    }

    basic_istream<charT, _traits>& operator>> (setw_t w)
    {
      this->width(w.width);
      return *this;
    }

  };

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& getline(basic_istream<charT, _traits>& is, DS_STRING_NAMESPACE::basic_string<charT>& str, const charT delim='\n')
  {
    do
      {
	const int c=is.get();
	if(static_cast<charT>(c)==delim || is.fail())
	  break;
	str+=static_cast<charT>(c);
      }
    while(true);
    return is;
  }

  template <TYPENAME charT, TYPENAME _traits>
  int basic_istream<charT, _traits>::getImp()
  {
    if(this->fail())
      return EOF;
    const int c=this->sbget();
    if(c<0)
      setstate(bios::eofbit);
    else
      count++;
    return c;
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_istream<charT, _traits>::skipwhitespace()
  {
    bool flag=true;
    do
      {
	const int c=getImp();
	if(c<0) flag=false;
	if(!isws(c))
	  {
	    unget();
	    flag=false;
	  }
      } while(flag && this->good());
  }

  template <TYPENAME charT, TYPENAME _traits>
  streampos basic_istream<charT, _traits>::tellg() const
  {
    streampos p=0;
    if(sb_)
      p=sb_->pubseekoff(0, ios::cur, ios::in);
    return p;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::seekg(const streamoff offs)
  {
    if(sb_)
      sb_->pubseekpos(offs, ios::in);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::seekg(const streamoff offs, const ios::seekdir flag)
  {
    if(sb_)
      sb_->pubseekoff(offs, flag, ios::in);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  int basic_istream<charT, _traits>::sync() const
  {
    if(sb_)
      return sb_->pubsync();
    return -1;
  }

  template <TYPENAME charT, TYPENAME _traits>
  int basic_istream<charT, _traits>::get()
  {
    count=0;
    return getImp();
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::get(charT& c)
  {
    count=0;
    c=static_cast<charT>(getImp());
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::get(charT* zf, size_t anz, charT ende)
  {
    count=0;
    while(anz-- && this->good())
      {
	const int i=getImp();
	if(i<0 || i==ende)
	  break;
	*zf++=static_cast<charT>(i);
      }
    *zf=0;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::getline(charT *zf, size_t anz, charT ende)
  {
    count=0;
    while(anz-- && this->good())
      {
	const int i=getImp();
	if(i<0)
	  break;
	*zf++=static_cast<charT>(i);
	if(i==ende)
	  break;
      }
    *zf=0;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::read(charT *zf, size_t anz)
  {
    count=0;
    while(anz-- && this->good())
      {
	const int i=getImp();
	if(i<0)
	  break;
	*zf++=static_cast<charT>(i);
      }
    if(this->rdstate()&bios::eofbit)
      setstate(bios::failbit);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::readsome(charT *zf, size_t anz)
  {
    count=0;
    while(anz-- && this->good())
      {
	const int i=getImp();
	if(i<0)
	  break;
	*zf++=static_cast<charT>(i);
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::ignore(int anz, int ende)
  {
    while(anz-- && this->good())
      if(getImp()<0)
	break;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  int basic_istream<charT, _traits>::peek()
  {
    basic_streambuf<charT> *s=this->rdbuf();
    return s?s->sgetc():EOF;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::unget()
  {
    basic_streambuf<charT> *s=this->rdbuf();
    if(s)
      {
	s->sungetc();
	--count;
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::putback(const charT c)
  {
    basic_streambuf<charT> *s=this->rdbuf();
    if(s)
      {
	s->sputbackc(c);
	--count;
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::get(basic_streambuf<charT>& s, const charT delim)
  {
    for(;;)
      {
	const int r=this->getImp();
	if(r<0)
	  break;
	if(static_cast<charT>(r)==delim)
	  {
	    unget();
	    break;
	  }
	const int rr=s.sputc(r);
	if(rr<0)
	  {
	    unget();
	    break;
	  }
      }

    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (bool& b)
  {
    // todo: make recognition of number better
    DS_STRING_NAMESPACE::basic_string<charT> s;
    operator>> (s);
    basic_ios_traits<charT> t;
    if(s==t.truestring)
      {
	b=true;
	return *this;
      }
    else if(s==t.falsestring || s==t.zerostring)
      {
	b=false;
	return *this;
      }

    b=true;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (unsigned int& num)
  {
    count=0;
    num=0;
    skipwhitespace();
    if(this->fail())
      return *this;

    int multiplier=10;
    int c=getImp();
    if(c<0)
      return *this;
    if(c=='+')
      c=getImp(); // '+' consumed
    if(c<0)
      return *this;
    // leading zero indicates either octal or hexadecimal
    if(c=='0')
      {
	// first set octal
	multiplier=8;
	c=getImp();
	if(c<0)
	  return *this;
	// check for hex
	if(c=='x' || c=='X')
	  multiplier=16;
	else
	  // no, no hex :-( you octal suckers
	  unget();
      }
    else
      // ok, we take the beloved 10
      unget();

    // read number
    do {
      c=getImp();
      if(c<0)
	return *this;
      if(isws(c))
	{
	  unget();
	  break;
	}
      // check valid characters along with base
      if(c>='0' && c<='7')
	c-='0';
      else if(c>='8' && c<='9' && multiplier<=10)
	c-='0';
      else if(c>='a' && c<='f' && multiplier<=16)
	c-='a'+10;
      else if(c>='A' && c<='F' && multiplier<=16)
	c-='A'+10;
      else
	break; // invalid character for this mode

      num*=multiplier	;
      num+=c;
    } while(this->good());

    // read until we reach whitespace again
    do {
      c=getImp();
    } while(c>=0 && (!isws(c)) && this->good());

    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (int& num)
  {
    count=0;
    num=0;
    skipwhitespace();
    if(this->fail())
      return *this;

    // check for '-'
    bool neg=false;
    const int c=getImp();
    if(c<0)
      return *this;
    if(c=='-')
      neg=true;
    else
      unget();

    unsigned int n;
    operator>> (n);

    num=static_cast<int>(n);
    if(neg)
      num=-num;

    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (charT *str)
  {
    count=0;
    skipwhitespace();

    int c;
    do {
      c=getImp();
      if(c<0)
	return *this;
      if(this->whitespace(c))
	break;
      *str++=static_cast<charT>(c);
    } while(this->good());

    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (DS_STRING_NAMESPACE::basic_string<charT>& str)
  {
    str.erase();
    count=0;
    skipwhitespace();

    int c;
    do {
      c=getImp();
      if(c<0)
	return *this;
      if(isws(c))
	break;
      str+=static_cast<charT>(c);
    } while(this->good());

    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (void* &v)
  {
    unsigned int i;
    operator>>(i);
    v=reinterpret_cast<void*>(i);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_istream<charT, _traits>& basic_istream<charT, _traits>::operator>> (double& d)
  {
    DS_STRING_NAMESPACE::basic_string<charT> s;
    if(*this >> s)
      {
	DS_STRING_NAMESPACE::string buf(s.size(), 0);
	narrow_(const_cast<char*>(buf.c_str()), const_cast<charT*>(s.c_str()));
	if(sscanf(buf.c_str(), "%lf", &d) != 1)
	  setstate(bios::failbit);
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  bool basic_istream<charT, _traits>::isws(const int c)
  {
#if defined(__GNUC__)
    if(c>=0 && c<256)
      return isspace(c);
    return false;
#else
    return isspace(c);
#endif
  }

  typedef basic_istream<char> istream;
  typedef basic_istream<wchar_t> wistream;
}

#endif
