// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef DS_BASICOSTREAM__HPP
#define DS_BASICOSTREAM__HPP

#include <iosfwd>
#include <ds_basicios.hpp>
#include <string>

namespace DS_NAMESPACE
{
  template <TYPENAME charT,
	    TYPENAME _traits
#if defined(__GNUC__)
	    = int
#endif
	    >
  class basic_ostream : public virtual basic_ios<charT>
  {
    typedef basic_ios<charT> bios;
    using bios::sb_;
  private:
    void putImp(const charT c);
    void putNum(const DS_STRING_NAMESPACE::basic_string<charT>& s);
    void putFill(const size_t ssize);
    void putStr(const DS_STRING_NAMESPACE::basic_string<charT>& s);
    void putUnsigned(unsigned int num, charT prefix);
    void fixAdjustFlags();

  public:
    basic_ostream<charT, _traits>(basic_streambuf<charT> *s=0) :
      basic_ios<charT>(s)
    {
      //ds_debughook("basic_ostream()");
    }

    ~basic_ostream<charT, _traits>()
    {
      //ds_debughook("~basic_ostream()");
    }

    streampos tellp() const;
    basic_ostream<charT, _traits>& seekp(const streampos p);
    basic_ostream<charT, _traits>& seekp(const streamoff offs, const ios::seekdir flag);
    basic_ostream<charT, _traits>& put(const charT c);

    basic_ostream<charT, _traits>& write(const charT* zf, size_t anz)
    {
      if(zf && anz && sb_)
	sb_->sputn(zf, anz);
      return *this;
    }

    basic_ostream<charT, _traits>& flush();

    basic_ostream<charT, _traits>& operator<< (const charT c)
    {
      return this->put(c);
    }


    basic_ostream<charT, _traits>& operator<< (unsigned int num)
    {
      this->putUnsigned(num, 0);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (const charT *str)
    {
      if(str && sb_)
	{
	  const charT *p=str;
	  unsigned z=0;
	  while(*p++) ++z;
	  if(z)
	    sb_->sputn(str, z);
	}
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (const bool b);
    basic_ostream<charT, _traits>& operator<< (int num);
    basic_ostream<charT, _traits>& operator<< (const DS_STRING_NAMESPACE::basic_string<charT>& str);
    basic_ostream<charT, _traits>& operator<< (const void *v);
    basic_ostream<charT, _traits>& operator<< (basic_streambuf<charT> *sb);

    basic_ostream<charT, _traits>& operator<< (double num);

    /// manipulator overload
    basic_ostream<charT, _traits>& operator<< ( basic_ostream<charT, _traits>& (*func) (basic_ostream<charT, _traits>&) )
    {
      return func(*this);
    }

    basic_ostream<charT, _traits>& operator<< ( ios_base& (*func) (ios_base&) )
    {
      func(*this);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< ( basic_ios<charT>& (*func) (basic_ios<charT>&) )
    {
      func(*this);
      return *this;
    }

    // iomanip overloads

    basic_ostream<charT, _traits>& operator<< (resetiosflags_t mask)
    {
      this->unsetf(mask.flags);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (setiosflags_t mask)
    {
      this->setf(mask.flags);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (setbase_t base)
    {
      switch(base.base)
	{
	case 8:  this->setf(bios::oct, bios::basefield); break;
	case 10: this->setf(bios::dec, bios::basefield); break;
	case 16: this->setf(bios::hex, bios::basefield); break;
	default: this->setf(0, bios::basefield); break;
	}
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (setfill_t f)
    {
      this->fill(f.fill);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (setprecision_t p)
    {
      this->precision(p.prec);
      return *this;
    }

    basic_ostream<charT, _traits>& operator<< (setw_t w)
    {
      this->width(w.width);
      return *this;
    }


  };

  typedef basic_ostream<char> ostream;
  typedef basic_ostream<wchar_t> wostream;

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::putImp(const charT c)
  {
    if(!this->good())
      return;

    const int i=this->sbput(c);
    if(i < 0)
      this->setstate(bios::failbit);
    else
      this->clear();
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::putNum(const DS_STRING_NAMESPACE::basic_string<charT>& s)
  {
    typename DS_STRING_NAMESPACE::basic_string<charT>::const_iterator i=s.end();
    for(--i; i>=s.begin() && this->good(); --i)
      this->putImp(*i);
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::putFill(const size_t ssize)
  {
    if(this->width_>static_cast<streamsize>(ssize))
      {
	this->width_-=ssize;
	while(this->width_--)
	  this->putImp(this->fill_);
      }
    this->width_=0;
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::putStr(const DS_STRING_NAMESPACE::basic_string<charT>& s)
  {
    if(!s.empty() && sb_)
      sb_->sputn(s.data(), s.size());
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::putUnsigned(unsigned int num, charT prefix)
  {
    const ios::fmtflags sav=this->fmtflags_;

    // fix fmtflags_
    const unsigned f=this->fmtflags_&bios::basefield;
    if(f==0 || f==(bios::oct|bios::dec) || f>bios::hex)
      this->setf(bios::dec, bios::basefield);
    this->fixAdjustFlags();

    typedef DS_STRING_NAMESPACE::basic_string<charT> s_t;
    s_t s;
    s_t prefixStr;
    if(prefix)
      prefixStr+=prefix;

    // shortcut for 0
    if(num==0)
      {
	charT buf[4]; widen(buf, "0"); s=buf;
      }
    else
      {
	// get character table
	basic_ios_traits<charT> traits;
	charT *tab=(this->fmtflags_&bios::uppercase)?traits.uppernumber:traits.lowernumber;

	// set divisor
	const int divisor=(this->fmtflags_&bios::hex)?16:( (this->fmtflags_&bios::oct)?8:10 );

	// and now the number
	do {
	  s+=tab[num%divisor];
	  num/=divisor;
	} while(num);

	// print base
	if(this->fmtflags_&bios::showbase)
	  {
	    if(divisor==8)
	      s+='0';
	    else if(divisor==16)
	      {
		prefixStr+='0';
		prefixStr+=tab[16]; // x
	      }
	  }

      }

    if(this->fmtflags_&bios::right)
      {
	this->putFill(s.size()+prefixStr.size());
	this->putStr(prefixStr);
	this->putNum(s);
      }
    else if(this->fmtflags_&bios::left)
      {
	this->putStr(prefixStr);
	this->putNum(s);
	this->putFill(s.size()+prefixStr.size());
      }
    else if(this->fmtflags_&bios::internal)
      {
	this->putStr(prefixStr);
	this->putFill(s.size()+prefixStr.size());
	this->putNum(s);
      }

    this->width_=0;
    this->fmtflags_=sav;
  }

  template <TYPENAME charT, TYPENAME _traits>
  void basic_ostream<charT, _traits>::fixAdjustFlags()
  {
    const unsigned f=this->fmtflags_&bios::adjustfield;
    if(f==0 || f==(bios::left|bios::right) || f>bios::internal)
      this->setf(bios::right, bios::adjustfield);
  }

  template <TYPENAME charT, TYPENAME _traits>
  streampos basic_ostream<charT, _traits>::tellp() const
  {
    streampos p=0;
    if(sb_)
      p=sb_->pubseekoff(0, ios::cur, ios::out);
    return p;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::seekp(const streampos p)
  {
    if(sb_)
      sb_->pubseekpos(p, ios::out);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::seekp(const streamoff offs, const ios::seekdir flag)
  {
    if(sb_)
      sb_->pubseekoff(offs, flag, ios::out);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::put(const charT c)
  {
    if(this->width_==0)
      this->putImp(c);
    else
      {
	const ios::fmtflags sav=this->fmtflags_;
	this->fixAdjustFlags();
	if(this->fmtflags_&bios::left)
	  this->putImp(c);
	while(--this->width_)
	  this->putImp(this->fill_);
	if((this->fmtflags_&bios::right) || (this->fmtflags_&bios::internal))
	  this->putImp(c);
	this->fmtflags_=sav;
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::flush()
  {
    if(sb_)
      sb_->pubsync();
    else
      this->setstate(bios::badbit);
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (const bool b)
  {
    if(this->fmtflags_&ios::boolalpha)
      {
	basic_ios_traits<charT> t;
	return this->operator<< (b?t.truestring:t.falsestring);
      }
    return this->put(b?'1':'0');
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (int num)
  {
    // oct and hex don't make negative output, they interpret themselfs as unsigned
    if(this->fmtflags_&bios::hex || this->fmtflags_&bios::oct)
      {
	this->putUnsigned(*reinterpret_cast<unsigned int*>(&num), 0);
      }
    // dec output
    else
      {
	charT prefix=0;
	if(num<0)
	  {
	    prefix='-';
	    num=-num;
	  }
	else if(num>=0 && (this->fmtflags_&bios::showpos))
	  prefix='+';
	this->putUnsigned(static_cast<unsigned int>(num), prefix);
      }
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (const DS_STRING_NAMESPACE::basic_string<charT>& str)
  {
    const ios::fmtflags sav=this->fmtflags_;
    this->fixAdjustFlags();

    if((this->fmtflags_&bios::right) || (this->fmtflags_&bios::internal))
      this->putFill(str.size());

    this->putStr(str);

    if(this->fmtflags_&bios::left)
      this->putFill(str.size());

    this->fmtflags_=sav;
    this->width_=0;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (const void *v)
  {
    const ios::fmtflags fmtsav=this->fmtflags_;
    this->fmtflags_|=bios::uppercase;
    this->fmtflags_|=bios::showbase;
    this->fmtflags_|=bios::hex;
    this->operator<< (reinterpret_cast<unsigned int>(v));
    this->fmtflags_=fmtsav;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (basic_streambuf<charT> *sb)
  {
    if(sb)
      while(this->good())
	{
	  const int c=sb->sgetc();
	  if(c<0)
	    break;
	  this->putImp(static_cast<charT>(c));
	}
    this->width_=0;
    return *this;
  }

  template <TYPENAME charT, TYPENAME _traits>
  basic_ostream<charT, _traits>& basic_ostream<charT, _traits>::operator<< (double num)
  {
    const int bufsize=64;
    // todo: output is not correct exept for scientific mode
    char buf[bufsize]; char *b=buf;

    if(this->isset(bios::showpos) && num>= 0)
      *b++='+';

    char fmt[bufsize];
    if(this->isset(bios::scientific))
      snprintf(fmt, sizeof(fmt), "%%%i.%i%c",
	       static_cast<int>(this->precision_),
	       static_cast<int>(this->precision_),
	       this->isset(bios::uppercase)?'E':'e');
    else
      snprintf(fmt, sizeof(fmt), "%%%c", this->isset(bios::uppercase)?'F':'f');

    snprintf(b, sizeof(buf)-1, fmt, num);

    charT BUF[bufsize]; widen(BUF, buf);
    return *this << BUF;
  }

  // manipulators

  template <TYPENAME charT, TYPENAME _traits>
  inline basic_ostream<charT, _traits>& flush(basic_ostream<charT, _traits>& os)
  {
    return os.flush();
  }

  template <TYPENAME charT, TYPENAME _traits>
  inline basic_ostream<charT, _traits>& endl(basic_ostream<charT, _traits>& os)
  {
    return os << '\n' << flush;
  }

  template <TYPENAME charT, TYPENAME _traits>
  inline basic_ostream<charT, _traits>& ends(basic_ostream<charT, _traits>& os)
  {
    return os << '\0';
  }

}

#endif
