// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef STREAMBUF__HPP
#define STREAMBUF__HPP

#include <ds_iosbase.hpp>
#include <stdio.h>

namespace DS_NAMESPACE
{

  template <TYPENAME charT>
  struct basic_streambuf_traits
  {
    typedef unsigned int unsigned_type;
  };

  template <>
  struct basic_streambuf_traits<char>
  {
    typedef unsigned char unsigned_type;
  };

  template <>
  struct basic_streambuf_traits<wchar_t>
  {
#if defined(_WIN32_WCE)
    typedef wchar_t unsigned_type;
#else
    typedef unsigned wchar_t unsigned_type;
#endif
  };


  template <TYPENAME charT>
  class basic_streambuf
  {
  public:
    typedef typename basic_streambuf_traits<charT>::unsigned_type char_type;
    typedef int int_type;
    typedef unsigned pos_type;
    typedef int off_type;

    virtual ~basic_streambuf();

    basic_streambuf* pubsetbuf(char_type *buffer, streamsize count) { return this->setbuf(buffer,count); }
    pos_type pubseekoff(off_type off, ios_base::seekdir way, ios_base::openmode which = ios_base::in | ios_base::out) { return this->seekoff(off,way,which); }
    pos_type pubseekpos(pos_type sp, ios_base::openmode which = ios_base::in | ios_base::out) { return this->seekpos(sp,which); }
    int pubsync() { return this->sync(); }

    streamsize in_avail()
    {
      if(gptr_)
	return egptr_-gptr_;
      return this->showmanyc();
    }

    int_type snextc()
    {
      return (this->sbumpc()==EOF)?EOF:this->sgetc();
    }
    int_type sbumpc()
    {
      if(gptr_ && gptr_<egptr_)
	return *gptr_++;
      return this->uflow();
    }
    int_type sgetc()
    {
      if(gptr_ && gptr_<egptr_)
	return *gptr_;
      return this->underflow();
    }

    void stossc() { this->sbumpc(); }

    streamsize sgetn(charT *ptr, streamsize count) { return this->xsgetn(ptr,count); }
    int_type sputbackc(char_type ch)
    {
      if(gptr_ && gptr_>eback_ && *(gptr_-1)==ch)
	return *(--gptr_)=ch;
      return this->pbackfail(ch);
    }
    int_type sungetc()
    {
      if(gptr_ && gptr_>eback_)
	return *(--gptr_);
      return this->pbackfail();
    }
    int_type sputc(char_type ch)
    {
      if(pptr_ && pptr_<epptr_)
	return *(++pptr_)=ch;
      return this->overflow(ch);
    }
    streamsize sputn(const charT *ptr, streamsize count) { return this->xsputn(ptr,count); }

  protected:
    basic_streambuf() :
      eback_(0),
      gptr_(0),
      egptr_(0),
      pbase_(0),
      pptr_(0),
      epptr_(0)
    { }

    char_type* eback() const { return eback_; }
    char_type* gptr() const { return gptr_; }
    char_type* egptr() const { return egptr_; }

    void gbump(int count);
    void setg(char_type *gbeg, char_type *gnext, char_type *gend)
    {
      eback_=gbeg;
      gptr_=gnext;
      egptr_=gend;
    }

    char_type* pbase() const { return pbase_; }
    char_type* pptr() const { return pptr_; }
    char_type* epptr() const { return epptr_; }

    void pbump(int count) { pptr+=count; }
    void setp(char_type *pbeg, char_type *pend)
    {
      pbase_=pptr_=pbeg;
      epptr_=pend;
    }

    virtual basic_streambuf* setbuf(char_type *buffer, streamsize count);

    virtual pos_type seekoff(off_type off, ios_base::seekdir way, ios_base::openmode which = ios_base::in | ios_base::out) { return EOF; }
    virtual pos_type seekpos(pos_type sp, ios_base::openmode which = ios_base::in | ios_base::out) { return EOF; }

    virtual streamsize showmanyc();

    virtual streamsize xsgetn(charT *ptr, streamsize count);

    virtual int_type underflow() { return EOF; }
    virtual int_type uflow()
    {
      int_type c=this->underflow();
      if(c!=EOF && gptr_<egptr_)
	++gptr_;
      return c;
    }

    virtual int_type pbackfail(int_type meta = EOF) { return EOF; }
    virtual streamsize xsputn(const charT *ptr, streamsize count);
    virtual int_type overflow(int_type c) { return EOF; }

    virtual int sync();

  private:
    char_type *eback_;
    char_type *gptr_;
    char_type *egptr_;
    char_type *pbase_;
    char_type *pptr_;
    char_type *epptr_;
  };

  template <TYPENAME charT>
  basic_streambuf<charT>::~basic_streambuf()
  { }

  template <TYPENAME charT>
  basic_streambuf<charT>* basic_streambuf<charT>::setbuf(basic_streambuf<charT>::char_type *buffer, streamsize count)
  {
    return this; 
  }

  template <TYPENAME charT>
  int basic_streambuf<charT>::sync()
  {
    return 0;
  }

  template <TYPENAME charT>
  streamsize basic_streambuf<charT>::showmanyc()
  {
    return 0;
  }

  template <TYPENAME charT>
  streamsize basic_streambuf<charT>::xsgetn(charT *ptr, streamsize count)
  {
    streamsize i=0;
    for(; i!=count; ++i)
      {
	int_type c=this->sbumpc();
	if(c==EOF)
	  break;
	*ptr++=static_cast<char_type>(c);
      }
    return i;
  }

  template <TYPENAME charT>
  streamsize basic_streambuf<charT>::xsputn(const charT *ptr, streamsize count)
  {
    streamsize i=0;
    for(; i!=count; ++i)
      {
	if(this->sputc(*ptr)==EOF)
	  break;
	++ptr;
      }
    return i;
  }


  typedef basic_streambuf<char> streambuf;
  typedef basic_streambuf<wchar_t> wstreambuf;

}

#endif
