// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#ifndef DS_BASICSTRINGBUF__HPP
#define DS_BASICSTRINGBUF__HPP

#include <ds_basicstreambuf.hpp>
#include <ds_basicios.hpp>
#include <string>

namespace DS_NAMESPACE
{

  template <TYPENAME charT>
  class basic_stringbuf : public basic_streambuf<charT>
  {
  public:
    typedef typename basic_streambuf<charT>::char_type char_type;
    typedef typename basic_streambuf<charT>::int_type int_type;
    typedef typename basic_streambuf<charT>::pos_type pos_type;
    typedef typename basic_streambuf<charT>::off_type off_type;

    basic_stringbuf(ios_base::openmode mode = ios_base::in | ios_base::out) :
      pos_(0),
      mode_(mode)
    { }

    basic_stringbuf(const DS_STRING_NAMESPACE::basic_string<charT>& str, ios_base::openmode mode = ios_base::in | ios_base::out) :
      pos_(0),
      mode_(mode)
    {
      if(mode_&ios::in)
	in_=str;
      if(mode_&ios::out)
	out_=str;
    }

    DS_STRING_NAMESPACE::basic_string<charT> str() const
    {
      if((mode_&ios::out) && !out_.empty())
	return out_;
      if((mode_&ios::in)  && !in_.empty())
	return in_;
      return DS_STRING_NAMESPACE::basic_string<charT>();
    }

    void str(const DS_STRING_NAMESPACE::basic_string<charT>& newstr)
    {
      if(mode_&ios::in)
	in_=newstr;
      if(mode_&ios::out)
	out_.erase();
    }

protected:
    virtual pos_type seekoff(off_type off, ios_base::seekdir way, ios_base::openmode mode = ios_base::in | ios_base::out)
    {
      pos_type ret=EOF;
      if(mode&ios::in)
	{
	  int_type i=pos_;
	  switch(way)
	    {
	    case ios::beg : i=off; break;
	    case ios::cur : i+=off; break;
	    case ios::end : i=in_.size()+off; break;
	    default: return EOF;
	    }
	  if(i>=0)
	    ret=pos_=i;
	}
      if(mode&ios::out)
	{
	  int_type i=pos_;
	  switch(way)
	    {
	    case ios::beg : i=off; break;
	    case ios::cur : i+=off; break;
	    case ios::end : i=out_.size()+off; break;
	    default: return EOF;
	    }
	  if(i>=0)
	    {
	      out_.resize(i);
	      ret=i;
	    }
	}
      return ret;
    }

    virtual pos_type seekpos(pos_type sp, ios_base::openmode mode = ios_base::in | ios_base::out)
    {
      // 31bit clipping. Will be fixed later
      return this->seekoff(sp, ios::beg, mode);
    }

    virtual int_type underflow()
    {
      if(pos_ < in_.size())
	return static_cast<char_type>(in_[pos_++]);
      return EOF;
    }

    virtual int_type pbackfail(int_type meta = EOF)
    {
      if(pos_>0)
	{
	  --pos_;
	  if(meta!=EOF)
	    in_[pos_]=meta;
	  return !EOF;
	}
      return EOF;
    }

    virtual int_type overflow(int_type meta = EOF)
    {
      if(meta!=EOF)
	out_+=static_cast<char_type>(meta);
      return meta;
    }

    virtual streamsize xsgetn(charT *ptr, streamsize count);
    virtual streamsize xsputn(const charT *ptr, streamsize count);

    DS_STRING_NAMESPACE::basic_string<charT> out_, in_;
    pos_type pos_;
    ios_base::openmode mode_;
    };

  template <TYPENAME charT>
  streamsize basic_stringbuf<charT>::xsgetn(charT *ptr, streamsize count)
  {
    streamsize z=0;
    while(pos_<in_.size() && count--)
      {
	*ptr++=in_[pos_++];
	++z;
      }
    return z;
  }

  template <TYPENAME charT>
  streamsize basic_stringbuf<charT>::xsputn(const charT *ptr, streamsize count)
  {
    out_.append(ptr, static_cast<size_t>(count));
    return count;
  }

}

#endif
