// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

// todo: fix the callback event stuff

#ifndef IOSBASE__HPP
#define IOSBASE__HPP

#include <ds_config.hpp>
#include <vector>

namespace DS_NAMESPACE
{

  typedef dojpos_t streampos;
  typedef dojoff_t streamoff;
  typedef int64_t streamsize;

  class ios_base
  {
  public:
    enum fmtflags
    {
      boolalpha = 0x01,
      showbase =  0x02,
      showpoint = 0x04,
      showpos =   0x08,
      skipws =    0x10,
      unitbuf =   0x20,
      uppercase = 0x40,

      left =	  0x100,
      right =	  0x200,
      internal =  0x400,
      adjustfield=left|right|internal,

      oct =	  0x1000,
      dec =	  0x2000,
      hex =	  0x4000,
      basefield = oct|dec|hex,

      fixed =	  0x10000,
      scientific= 0x20000,
      floatfield= fixed|scientific
    };

    enum openmode
    {
      in=1,
      out=2,
      app=4,
      ate=8,
      trunc=16,
      binary=32,
    };

    enum seekdir { beg, cur, end };

    enum iostate
    {
      goodbit = 1,
      eofbit =  2,
      failbit = 4,
      badbit =  8
    };

    enum event
      {
	copyfmt_event,
	erase_event,
	imbue_event
      };

  private:
    iostate state_;
#if 0
    iostate exceptionmask_;
#endif

  protected:
    streamsize width_;
    streamsize precision_;

    fmtflags fmtflags_;

    bool isset(fmtflags f)
    {
      return fmtflags_&f;
    }

  public:

    struct Init
    {
      Init();
      ~Init();
      static int count;
    };

    ios_base() :
      state_(goodbit),
      width_(0),
      precision_(6),
      fmtflags_(static_cast<fmtflags>(skipws|right|dec))
    { }

    virtual ~ios_base();

    int good() { return this->rdstate()&goodbit; }
    int eof() { return this->rdstate()&eofbit; }
    int fail() { return this->rdstate()&(failbit|badbit); }
    int bad() { return this->rdstate()&badbit; }

    operator void* () { return this->good()?this:0; }
    int operator! () { return this->fail(); }

    void clear(iostate state = goodbit) { this->setstate(state); }
    void setstate(iostate state) { state_=state; }
    iostate rdstate() const { return state_; }

    void exception(bool flags) { } // we don't support exceptions in dojstream
    bool exception() { return false; }

    fmtflags flags() const { return fmtflags_; }
    fmtflags flags(fmtflags n)
    {
      const fmtflags of=fmtflags_;
      fmtflags_=n;
      return of;
    }

    fmtflags setf(fmtflags n)
    { 
      const fmtflags of=fmtflags_;
      fmtflags_=static_cast<fmtflags>(fmtflags_|n);
      return of;
    }
    fmtflags setf(fmtflags n, fmtflags mask)
    {
      const fmtflags of=fmtflags_;
      fmtflags_= static_cast<fmtflags>((mask & n) | (of & (~mask)));
      return of;
    }
    void unsetf(fmtflags mask)
    {
      fmtflags_ = static_cast<fmtflags>(fmtflags_ & (~mask));
    }

    void copyfmt(const ios_base& i);

    streamsize precision() const { return precision_; }
    streamsize precision(streamsize np)
    {
      const streamsize op=precision_;
      precision_=np;
      return op;
    }

    streamsize width() const { return width_; }
    streamsize width(streamsize nw)
    {
      const streamsize ow=width_;
      width_=nw;
      return ow;
    }

    static bool sync_with_stdio(bool sync=true);
    static int xalloc();
    void*& pword(int idx);
    long&  iword(int idx);

    typedef void *(event_callback(event ev, ios_base& iosbase, int idx));
#if !defined(_WIN32_WCE)
    void register_callback(event_callback pfn, int idx);
#endif
  private:
#if 0
    std::vector<event_callback> ecb_vec;
    std::vector<int>            idx_vec;
#endif
    void do_callbacks(event e);

    std::vector<long> long_vec;
    std::vector<void*> void_vec;
  };

  // manipulators
  inline ios_base& boolalpha(ios_base& i)   { i.setf(ios_base::boolalpha); return i; }
  inline ios_base& noboolalpha(ios_base& i) { i.unsetf(ios_base::boolalpha); return i; }

  inline ios_base& showbase(ios_base& i)   { i.setf(ios_base::showbase); return i; }
  inline ios_base& noshowbase(ios_base& i) { i.unsetf(ios_base::showbase); return i; }

  inline ios_base& showpoint(ios_base& i)   { i.setf  (ios_base::showpoint); return i; }
  inline ios_base& noshowpoint(ios_base& i) { i.unsetf(ios_base::showpoint); return i; }

  inline ios_base& showpos(ios_base& i)   { i.setf  (ios_base::showpos); return i; }
  inline ios_base& noshowpos(ios_base& i) { i.unsetf(ios_base::showpos); return i; }

  inline ios_base& skipws(ios_base& i)   { i.setf  (ios_base::skipws); return i; }
  inline ios_base& noskipws(ios_base& i) { i.unsetf(ios_base::skipws); return i; }

  inline ios_base& unitbuf(ios_base& i)   { i.setf  (ios_base::unitbuf); return i; }
  inline ios_base& nounitbuf(ios_base& i) { i.unsetf(ios_base::unitbuf); return i; }

  inline ios_base& uppercase(ios_base& i)   { i.setf  (ios_base::uppercase); return i; }
  inline ios_base& nouppercase(ios_base& i) { i.unsetf(ios_base::uppercase); return i; }

  inline ios_base& left(ios_base& i)     { i.setf(ios_base::left, ios_base::adjustfield); return i; }
  inline ios_base& right(ios_base& i)    { i.setf(ios_base::right, ios_base::adjustfield); return i; }
  inline ios_base& internal(ios_base& i) { i.setf(ios_base::internal, ios_base::adjustfield); return i; }

  inline ios_base& oct(ios_base& i) { i.setf(ios_base::oct, ios_base::basefield); return i; }
  inline ios_base& dec(ios_base& i) { i.setf(ios_base::dec, ios_base::basefield); return i; }
  inline ios_base& hex(ios_base& i) { i.setf(ios_base::hex, ios_base::basefield); return i; }

  inline ios_base& fixed(ios_base& i)      { i.setf(ios_base::fixed, ios_base::floatfield); return i; }
  inline ios_base& scientific(ios_base& i) { i.setf(ios_base::scientific, ios_base::floatfield); return i; }

}

#include <ds_iomaniptypes.hpp>

#endif
