// dojstreams - a c++ stream library implementation
//
// Copyright (c) 2003-2004 by Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

// todo:
//   - replace methods
//   - getline() function
//   - check all functions against standard

#ifndef DS_STRING__HPP
#define DS_STRING__HPP

#include <ds_config.hpp>

namespace DS_STRING_NAMESPACE
{

#ifdef STRING_EXCEPTION
  struct out_of_range { };
  struct length_error { };
#endif

  template <TYPENAME charT>
  class basic_string_reverse_iterator
  {
  protected:
    charT* data_;
    int pos;
  public:
    basic_string_reverse_iterator<charT>(charT* d, int p)
      : data_(d), pos(p)
    { }

    charT* base() const
    {
      return &data_[pos];
    }

    void swap(basic_string_reverse_iterator<charT>& i)
    {
      charT* d=data_; data_=i.data_; i.data_=d;
      int p=pos; pos=i.pos; i.pos=p;
    }

    bool operator== (const basic_string_reverse_iterator<charT>& i)
    {
      return data_==i.data_ && pos==i.pos;
    }

    bool operator!= (const basic_string_reverse_iterator<charT>& i)
    {
      return !(*this==i);
    }

    bool operator<  (const basic_string_reverse_iterator<charT>& i)
    {
      return pos>i.pos;
    }

    bool operator>  (const basic_string_reverse_iterator<charT>& i)
    {
      return (!(*this<i)) && (!(*this==i));
    }

    bool operator<= (const basic_string_reverse_iterator<charT>& i)
    {
      return (*this<i) || (*this==i);
    }

    bool operator>= (const basic_string_reverse_iterator<charT>& i)
    {
      return !(*this<i);
    }

    basic_string_reverse_iterator<charT>& operator++()
    {
      --pos;
      return *this;
    }

    basic_string_reverse_iterator<charT> operator++(int)
    {
      basic_string_reverse_iterator<charT> i(*this);
      --pos;
      return i;
    }

    basic_string_reverse_iterator<charT>& operator--()
    {
      ++pos;
      return *this;
    }

    basic_string_reverse_iterator<charT> operator--(int)
    {
      basic_string_reverse_iterator<charT> i(*this);
      ++pos;
      return i;
    }

    basic_string_reverse_iterator<charT>& operator+= (int i)
    {
      pos+=i;
      return *this;
    }

    basic_string_reverse_iterator<charT>& operator-= (int i)
    {
      pos-=i;
      return *this;
    }

    basic_string_reverse_iterator<charT> operator+ (int i) const
    {
      return basic_string_reverse_iterator<charT>(*this)+=i;
    }

    basic_string_reverse_iterator<charT> operator- (int i) const
    {
      return basic_string_reverse_iterator<charT>(*this)-=i;
    }

    int operator- (const basic_string_reverse_iterator<charT>& i) const
    {
      return i.pos-pos;
    }

    basic_string_reverse_iterator<charT>& advance(int i)
    {
      pos+=i;
      return *this;
    }

    charT& operator* () const
    {
      return data_[pos];
    }

    charT* operator-> () const
    {
      return &data_[pos];
    }

  };


  template <TYPENAME charT>
  class basic_string
  {
  public:

    typedef unsigned int size_type;

#if defined(_MSC_VER)
    enum { npos = ~0 };
#else
    static const size_type npos = ~0;
#endif

#define __BASIC_STRING_MAX 0x7fffffff

    typedef charT* iterator;
    typedef const charT* const_iterator;

    typedef basic_string_reverse_iterator<charT> reverse_iterator;
    typedef reverse_iterator const_reverse_iterator;

  protected:

    charT* data_;
    size_type size_;
    size_type allocated_;


    void constructor();
    size_type size(const charT* str) const
    {
      size_type s=0;
      while(*str++) ++s;
      return s;
    }


  public:

    // constructors
    basic_string<charT>()
    {
      constructor();
    }

    basic_string<charT>(const basic_string<charT>& str, size_type idx=0, size_type num=__BASIC_STRING_MAX);
    basic_string<charT>(const charT* str, size_type num=__BASIC_STRING_MAX);

    basic_string<charT>(size_type num, charT c)
    {
      constructor();
      append(num, c);
    }

    ~basic_string<charT>();


    // access functions
    charT& operator[] (size_type idx) { return data_[idx]; }
    charT operator[] (size_type idx) const { return data_[idx]; }

    charT& at (size_type idx)
    {
      if(idx<size())
	return data_[idx];
#ifdef STRING_EXCEPTION
      throw out_of_range();
#endif
      return data_[0];
    }

    const charT* c_str() const { data_[size_]=0; return data_; }
    const charT* data() const { return data_; }

    // assign functions

    basic_string<charT>& assign(const basic_string<charT>& str, size_type idx, size_type anz);
    basic_string<charT>& assign(const basic_string<charT>& str)
    {
      return assign(str, 0, __BASIC_STRING_MAX);
    }
    basic_string<charT>& operator=(const basic_string<charT>& str)
    {
      return assign(str);
    }

    basic_string<charT>& assign(const charT* str, size_type anz);
    basic_string<charT>& assign(const charT* str)
    {
      return assign(str, __BASIC_STRING_MAX);
    }
    basic_string<charT>& operator=(const charT* str)
    {
      return assign(str);
    }

    basic_string<charT>& assign(size_type anz, charT c);
    basic_string<charT>& operator= (charT c)
    {
      data_[0]=c;
      size_=1;
      return *this;
    }

    void swap(basic_string<charT>& str)
    {
      charT* d=data_; data_=str.data_; str.data_=d;
      size_type s=size_; size_=str.size_; str.size_=s;
      size_type a=allocated_; allocated_=str.allocated_; str.allocated_=a;
    }

    // append functions

    basic_string<charT>& append(const basic_string<charT>& str, size_type idx, size_type anz);
    basic_string<charT>& append(const basic_string<charT>& str)
    {
      return append(str, 0, __BASIC_STRING_MAX);
    }
    basic_string<charT>& operator+= (const basic_string<charT>& str)
    {
      return append(str);
    }

    basic_string<charT>& append(const charT* str, size_type anz);
    basic_string<charT>& append(const charT* str)
    {
      return append(str, __BASIC_STRING_MAX);
    }
    basic_string<charT>& operator+= (const charT* str)
    {
      return append(str);
    }

    basic_string<charT>& append (size_type anz, charT c);
    basic_string<charT>& operator+= (const charT c)
    {
      return append(1, c);
    }

    void push_back(const charT c)
    {
      operator+=(c);
    }

    // query functions

    size_type max_size() const { return __BASIC_STRING_MAX; }
    size_type size() const { return size_; }
    size_type length() const { return size_; }
    size_type capacity() const { return allocated_; }

    bool empty() const { return !size(); }

    // compare functions

    int compare(const basic_string<charT>& str, size_type idx=0, size_type num=__BASIC_STRING_MAX) const;
    int compare(const charT* str, size_type idx=0, size_type num=__BASIC_STRING_MAX) const;

    // resize and reserve

    void reserve(size_type anz);
    void resize(size_type anz, charT c=' ');

    // erase functions

    basic_string<charT>& clear()
    {
      size_=0;
      return *this;
    }

    basic_string<charT>& erase(size_type idx=0, size_type anz=npos);

    // copy functions

    size_type copy(charT* str, size_type anz, size_type idx=0) const
    {
      if(idx>size())
	{
#ifdef STRING_EXCEPTION
	  throw out_of_range();
#else
	  return 0;
#endif
	}

      size_type i=0;
      for(iterator it=begin(); it!=end() && i<anz; ++it, ++i)
	*str++=*it;

      return i;
    }

    // substr

    basic_string<charT> substr(size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return basic_string<charT>(*this, idx, anz);
    }

    // iterator stuff

    iterator begin() const { return data_; }
    iterator end() const   { return data_+size(); }

    reverse_iterator rbegin() const { return reverse_iterator(data_, size()-1); }
    reverse_iterator rend() const   { return reverse_iterator(data_, -1); }

    template <TYPENAME Iter>
    basic_string<charT>& append(Iter begin, const Iter& end)
    {
      while(begin != end)
	{
	  operator+=(*begin);
	  ++begin;
	}
      return *this;
    }

    // find

    // todo: move outside class
    size_type find(const charT* str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      const size_type strsize=size(str);
      while(idx+strsize <= size())
	{
	  size_type i;
	  for(i=0; i<strsize && i<anz; ++i)
	    if(data_[idx+i]!=str[i])
	      {
		i=npos;
		break;
	      }
	  if(i!=npos)
	    return idx;
	  ++idx;
	}
      return npos;
    }

    size_type rfind(const charT* str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      // todo: implement this function
      return npos;
    }

    size_type find(const basic_string<charT>& str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return find(str.c_str(), idx, anz);
    }

    size_type rfind(const basic_string<charT>& str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return rfind(str.c_str(), idx, anz);
    }

    // todo: move outside class
    size_type find(const charT c, size_type idx=0) const
    {
      while(idx<size())
	{
	  if(data_[idx]==c)
	    return idx;
	  ++idx;
	}
      return npos;
    }

    // todo: move outside class
    size_type rfind(const charT c, size_type idx=0) const
    {
      if(size()==0)
	return npos;
      if(idx>=size())
	idx=size()-1;
      while(idx>=0)
	{
	  if(data_[idx]==c)
	    return idx;
	  --idx;
	}
      return npos;
    }

    // find first of

    // todo: move outside class
    size_type find_first_of(const charT* str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      size_type strsize=size(str);
      if(anz>strsize)
	anz=strsize;
      else if(strsize>anz)
	strsize=anz;

      while(idx<size())
	{
	  for(size_type i=0; i<strsize; i++)
	    if(data_[idx]==str[i])
	      return idx;
	  ++idx;
	}
      return npos;
    }

    size_type find_first_of(const basic_string<charT>& str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return find_first_of(str.c_str(), idx, anz);
    }

    size_type find_first_of(const charT c, size_type idx=0) const
    {
      return find(c, idx);
    }

    // find first not of

    // todo: move outside class
    size_type find_first_not_of(const charT* str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      size_type strsize=size(str);
      if(anz>strsize)
	anz=strsize;
      else if(strsize>anz)
	strsize=anz;

      while(idx<size())
	{
	  for(size_type i=0; i<strsize; ++i)
	    if(data_[idx]!=str[i])
	      return idx;
	  ++idx;
	}

      return npos;
    }

    size_type find_first_not_of(const basic_string<charT>& str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return find_first_not_of(str.c_str(), idx, anz);
    }

    // todo: move outside class
    size_type find_first_not_of(const charT c, size_type idx=0) const
    {
      while(idx<size())
	{
	  if(data_[idx]!=c)
	    return idx;
	  ++idx;
	}
      return npos;
    }


    // find last of

    // todo: move outside class
    size_type find_last_of(const charT* str, size_type idx=__BASIC_STRING_MAX, size_type anz=__BASIC_STRING_MAX) const
    {
      if(size()==0)
	return npos;

      size_type strsize=size(str);
      if(anz>strsize)
	anz=strsize;
      else if(strsize>anz)
	strsize=anz;

      if(idx>=size())
	idx=size()-1;

      while(idx>=0)
	{
	  for(size_type i=0; i<strsize; i++)
	    if(data_[idx]==str[i])
	      return idx;
	  --idx;
	}
      return npos;
    }

    size_type find_last_of(const basic_string<charT>& str, size_type idx=__BASIC_STRING_MAX, size_type anz=__BASIC_STRING_MAX) const
    {
      return find_last_of(str.c_str(), idx, anz);
    }

    // todo: move outside class
    size_type find_last_of(const charT c, size_type idx=__BASIC_STRING_MAX) const
    {
      if(size()==0)
	return npos;

      if(idx>=size())
	idx=size()-1;
      while(idx>=0)
	{
	  if(data_[idx]==c)
	    return idx;
	}
      return npos;
    }


    // find last not of

    // todo: move outside class
    size_type find_last_not_of(const charT* str, size_type idx=__BASIC_STRING_MAX, size_type anz=__BASIC_STRING_MAX) const
    {
      if(size()==0)
	return npos;

      size_type strsize=size(str);
      if(anz>strsize)
	anz=strsize;
      else if(strsize>anz)
	strsize=anz;

      if(idx>=size())
	idx=size()-1;

      while(idx>=0)
	{
	  for(size_type i=0; i<strsize; ++i)
	    if(data_[idx]!=str[i])
	      return idx;
	  --idx;
	}

      return npos;
    }

    size_type find_last_not_of(const basic_string<charT>& str, size_type idx=0, size_type anz=__BASIC_STRING_MAX) const
    {
      return find_last_not_of(str.c_str(), idx, anz);
    }

    // todo: move outside class
    size_type find_last_not_of(const charT c, size_type idx=__BASIC_STRING_MAX) const
    {
      if(size()==0)
	return npos;
      if(idx>=size())
	idx=size()-1;
      while(idx>=0)
	{
	  if(data_[idx]!=c)
	    return idx;
	  --idx;
	}
      return npos;
    }

  };

  template <TYPENAME charT>
  void basic_string<charT>::constructor()
  {
    size_=0;
    allocated_=4;
    data_=new charT[allocated_];
  }

  template <TYPENAME charT>
  basic_string<charT>::basic_string<charT>(const basic_string<charT>& str, size_type idx, size_type num)
    : data_(0)
  {
    basic_string<charT> s;
    s.append(str, idx, num);
    swap(s);
  }

  template <TYPENAME charT>
  basic_string<charT>::basic_string<charT>(const charT* str, size_type num)
    : data_(0)
  {
    basic_string<charT> s;
    s.append(str, num);
    swap(s);
  }

  template <TYPENAME charT>
  basic_string<charT>::~basic_string()
  {
    delete [] data_;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::assign(const basic_string<charT>& str, size_type idx, size_type anz)
  {
    basic_string<charT> s(str, idx, anz);
    swap(s);
    return *this;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::assign(const charT* str, size_type anz)
  {
    basic_string<charT> s(str, anz);
    swap(s);
    return *this;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::assign(size_type anz, charT c)
  {
    basic_string<charT> s(anz, c);
    swap(s);
    return *this;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::append(const basic_string<charT>& str, size_type idx, size_type anz)
  {
    const size_type strsize=str.size();
    if(strsize+size() > __BASIC_STRING_MAX)
      {
#ifdef STRING_EXCEPTION
	throw out_of_range();
#endif
	return *this;
      }

    // todo optimize here
    size_type i=idx, ii=0;
    while(i<strsize && ii<anz)
      {
	operator+=(str[i]);
	++i;
	++ii;
      }

    return *this;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::append(const charT* str, size_type anz)
  {
    if(str==0)
      return *this;

    const unsigned strsize=(anz==__BASIC_STRING_MAX)?size(str):anz;
    if(strsize+size() > __BASIC_STRING_MAX)
      {
#ifdef STRING_EXCEPTION
	throw out_of_range();
#endif
	return *this;
      }

    // todo: optimize this into a memcpy onto an expanded string
    for(unsigned i=0; i!=strsize; ++i)
      operator+=(*str++);

    return *this;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::append (size_type anz, charT c)
  {
    if(size()+anz > __BASIC_STRING_MAX)
      {
#ifdef STRING_EXCEPTION
	throw out_of_range();
#endif
	return *this;
      }

    resize(anz+size_, c);

    return *this;
  }

  template <TYPENAME charT>
  int basic_string<charT>::compare(const basic_string<charT>& str, size_type idx, size_type num) const
  {
    if(idx>size())
      return -2;

    size_type strI=0;
    while(idx<size() && strI<str.size() && strI<num)
      {
	if(data_[idx]<str.data_[strI])
	  return -1;
	if(data_[idx]>str.data_[strI])
	  return 1;
	++idx;
	++strI;
      }

    return 0;
  }

  template <TYPENAME charT>
  int basic_string<charT>::compare(const charT* str, size_type idx, size_type num) const
  {
    if(idx>size())
      return -2;

    size_type strI=0;
    while(idx<size() && *str && strI<num)
      {
	if(data_[idx]<*str)
	  return -1;
	if(data_[idx]>*str)
	  return 1;
	++idx;
	++strI;
	++str;
      }

    return 0;
  }

  template <TYPENAME charT>
  void basic_string<charT>::reserve(size_type anz)
  {
    if(anz<size())
      return;
    if(anz>__BASIC_STRING_MAX)
      anz=__BASIC_STRING_MAX;

    if(anz>allocated_)
      {
	while(allocated_<anz)
	  allocated_+=allocated_;
	charT* d=new charT[allocated_];
	charT* dest=d;
	const iterator e=end();
	for(iterator i=begin(); i!=e; ++i)
	  *dest++=*i;
	delete [] data_;
	data_=d;
      }
  }

  template <TYPENAME charT>
  void basic_string<charT>::resize(size_type anz, charT c)
  {
    reserve(anz+1); // reserve a charactor for \0
    if(anz<size())
      {
	size_=anz;
	return;
      }
    while(size_<anz)
      data_[size_++]=c;
  }

  template <TYPENAME charT>
  basic_string<charT>& basic_string<charT>::erase(size_type idx, size_type anz)
  {
    if(idx==0 && anz==npos)
      return clear();

    if(idx>size())
      {
#ifdef STRING_EXCEPTION
	throw out_of_range();
#else
	idx=size();
#endif
      }

    if(anz==npos)
      {
	resize(idx);
	return *this;
      }

    if(anz>=size())
      return clear();

    size_type i;
    for(i=0; i!=anz && idx+anz+i<size(); ++i)
      data_[idx+i]=data_[idx+anz+i];

    size_=idx+i;
    return *this;
  }



  // operator +

  template <TYPENAME charT>
  basic_string<charT> operator+ (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return basic_string<charT>(l)+=r;
  }

  template <TYPENAME charT>
  basic_string<charT> operator+ (const basic_string<charT>& l, const charT* r)
  {
    return basic_string<charT>(l)+=r;
  }

  template <TYPENAME charT>
  basic_string<charT> operator+ (const charT* l, const basic_string<charT>& r)
  {
    return basic_string<charT>(l)+=r;
  }

  template <TYPENAME charT>
  basic_string<charT> operator+ (const basic_string<charT>& l, const charT c)
  {
    return basic_string<charT>(l)+=c;
  }

  // operator ==

  template <TYPENAME charT>
  bool operator== (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return l.compare(r) == 0;
  }

  template <TYPENAME charT>
  bool operator== (const basic_string<charT>& l, const charT* r)
  {
    return l.compare(r) == 0;
  }

  template <TYPENAME charT>
  bool operator== (const charT* l, const basic_string<charT>& r)
  {
    return r.compare(l) == 0;
  }

  // operator !=

  template <TYPENAME charT>
  bool operator!= (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return !(l==r);
  }

  template <TYPENAME charT>
  bool operator!= (const basic_string<charT>& l, const charT* r)
  {
    return !(l==r);
  }

  template <TYPENAME charT>
  bool operator!= (const charT* l, const basic_string<charT>& r)
  {
    return !(l==r);
  }

  // operator <

  template <TYPENAME charT>
  bool operator< (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return l.compare(r) < 0;
  }

  template <TYPENAME charT>
  bool operator< (const basic_string<charT>& l, const charT* r)
  {
    return l.compare(r) < 0;
  }

  template <TYPENAME charT>
  bool operator< (const charT* l, const basic_string<charT>& r)
  {
    return r.compare(l) >= 0;
  }

  // operator >

  template <TYPENAME charT>
  bool operator> (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return l.compare(r) > 0;
  }

  template <TYPENAME charT>
  bool operator> (const basic_string<charT>& l, const charT* r)
  {
    return l.compare(r) > 0;
  }

  template <TYPENAME charT>
  bool operator> (const charT* l, const basic_string<charT>& r)
  {
    return r.compare(l) < 0;
  }

  // operator <=

  template <TYPENAME charT>
  bool operator<= (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return l.compare(r) <= 0;
  }

  template <TYPENAME charT>
  bool operator<= (const basic_string<charT>& l, const charT* r)
  {
    return l.compare(r) <= 0;
  }

  template <TYPENAME charT>
  bool operator<= (const charT* l, const basic_string<charT>& r)
  {
    return r.compare(l) > 0;
  }

  // operator >=

  template <TYPENAME charT>
  bool operator>= (const basic_string<charT>& l, const basic_string<charT>& r)
  {
    return l.compare(r) >= 0;
  }

  template <TYPENAME charT>
  bool operator>= (const basic_string<charT>& l, const charT* r)
  {
    return l.compare(r) >= 0;
  }

  template <TYPENAME charT>
  bool operator>= (const charT* l, const basic_string<charT>& r)
  {
    return r.compare(l) < 0;
  }


  template <TYPENAME charT>
  basic_string_reverse_iterator<charT> operator+ (const int n, const basic_string_reverse_iterator<charT>& i)
  {
    return basic_string_reverse_iterator<charT>(i) += n;
  }

  template <TYPENAME charT>
  void advance(basic_string_reverse_iterator<charT>& it, int n)
  {
    it+=n;
  }

  template <TYPENAME charT>
  int distance(const basic_string_reverse_iterator<charT>& a, const basic_string_reverse_iterator<charT>& b)
  {
    return a-b;
  }

  template <TYPENAME charT>
  void iter_swap(basic_string_reverse_iterator<charT>& a, basic_string_reverse_iterator<charT>& b)
  {
    a.swap(b);
  }


  template <TYPENAME charT>
  void swap(basic_string<charT>& l, basic_string<charT>& r)
  {
    l.swap(r);
  }

  typedef basic_string<char> string;
  typedef basic_string<wchar_t> wstring;
}

#endif
